#!/usr/bin/env python
# -*- coding: utf-8 -*-
def run():
	print("run")
	main()
	return None

def aaa():
	print("aaa")
	return None
	
def main():#
	print ("Hello, test is running")


if __name__ == '__main__':
    main()
